from tag_a_day.services.service import Service


class S3TagHandler(Service):
    name = 's3'
    missing_tags_text = "This bucket is missing '{0}' in its tags."

    def resources(self, session):
        return session.resource('s3').buckets.all()

    def handler(self, bucket, expected_tags, region, session, cache, proposals):
        try:
            tags = bucket.Tagging().tag_set
        except Exception:
            tags = []

        # Build table for displaying instance information
        evaluated_tags = self._progress.evaluated_tags(bucket.name)
        bucket_info, missing_tags = \
            self._build_tag_sets(expected_tags, evaluated_tags, tags)

        if self._progress.has_finished(bucket.name, expected_tags):
            self._skip(bucket.name)
            return

        if any(missing_tags):
            self._print_table(
                ("BucketName", bucket.name),
                *bucket_info
            )

            tag_prompt = self._build_tag_prompt(missing_tags)
            for tag_key in missing_tags:
                yield {
                    'resource_id': bucket.name,
                    'tag_key': tag_key,
                    'tag_value': tag_prompt(tag_key)
                }



from tag_a_day.services.service import Service


class EMRTagHandler(Service):
    name = 'emr_cluster'
    missing_tags_text = "This EMR cluster is missing '{0}' in its tags."

    def resources(self, session):
        emr = session.client('emr')
        for emr_page in emr.get_paginator('list_clusters').paginate():
            for cluster_summary in self._random_choose(emr_page['Clusters']):
                cluster = emr.describe_cluster(
                    ClusterId=cluster_summary['Id']).get('Cluster')
                yield cluster

    def handler(self, cluster, expected_tags, region, session, cache, proposals):
        # Get VPC info
        subnet = self._cache.subnet(
            cluster['Ec2InstanceAttributes']['Ec2SubnetId'],
            region
        )
        _, vpc_name, vpc = self._get_vpc_info(
            vpc_id=subnet.vpc_id,
            region=region
        )

        # Build table for displaying instance information
        evaluated_tags = self._progress.evaluated_tags(cluster['Id'])
        instance_info, missing_tags = \
            self._build_tag_sets(expected_tags, evaluated_tags, cluster['Tags'])

        if self._progress.has_finished(cluster['Id'], expected_tags):
            self._skip(cluster['Id'])
            return

        if any(missing_tags):
            self._print_table(
                ("Vpc", vpc_name),
                ("VpcID", vpc.id),
                ("ClusterID", cluster['Id']),
                *instance_info
            )

            tag_prompt = self._build_tag_prompt(missing_tags)
            for tag_key in missing_tags:
                yield {
                    'resource_id': cluster['Id'],
                    'tag_key': tag_key,
                    'tag_value': tag_prompt(tag_key),
                }



import abc
from operator import itemgetter

from prompt_toolkit import prompt
from tabulate import tabulate

from tag_a_day.cache import ProgressCache
from tag_a_day.log import logger


class Service(object):
    name = None
    prompt_text = "Propose a value for the '{0}' tag: "
    skip_text = "Skipping '{0}' as it has been evaluated\n\n"
    missing_tags_text = ''

    def __init__(self, session, table, cache):
        self._session = session
        self._proposals = table
        self._cache = cache
        self._proposer = ''
        self._progress = None

    def load(self):
        self._proposer = self._session().client('sts').get_caller_identity().get('Arn')
        self._progress = ProgressCache(self._proposals, self._proposer, self.name).load()
        return self

    def __repr__(self):
        return self.name

    def _get_vpc_info(self, vpc_id, region):
        # Get VPC information
        vpc = self._cache.vpc(vpc_id, region)
        vpc_name = [t for t in vpc.tags if t['Key'] == 'Name'][0]['Value']
        return vpc.id, vpc_name, vpc

    def _build_tag_sets(self, expected_tags, evaluated_tags, tags):
        instance_info = [('Tag.' + t['Key'], t['Value']) for t in tags]
        instance_info.sort(key=itemgetter(0))

        # Compute missing tags
        instance_tags = [tag['Key'] for tag in tags] + evaluated_tags
        missing_tags = [
            tag for tag in expected_tags if (tag not in instance_tags)]

        return instance_info, missing_tags

    def _random_choose(self, items):
        return items
        # for choice in random.choices(items, k=floor(len(items) * 0.6)):
        #     yield choice

    def _print_table(self, *args):
        print(tabulate(args))

    def _skip(self, resource_id):
        print(self.skip_text.format(resource_id))

    def _build_tag_prompt(self, missing_tags):
        print(self.missing_tags_text.format(
            "','".join(missing_tags)))

        # Build padding for easier to read prompt
        longest_key = len(max(missing_tags, key=len))
        justify_length = len(self.prompt_text) + longest_key

        def tag_prompt(tag_key):
            response = prompt(self.prompt_text.format(
                tag_key).ljust(justify_length))
            if len(response) < 1:
                print("Tag must not be empty")
                return tag_prompt(tag_key)
            else:
                return response

        return tag_prompt

    def run(self, expected_tags, region, session):
        logger().info("Auditing in region: {region}".format(region=region))
        with self._proposals.batch_writer() as proposals:
            for resource in self.resources(session):
                handled_tag = False
                for item in self.handler(resource, expected_tags, region, session, self._cache, proposals):
                    handled_tag = True
                    item['resource_type'] = self.name
                    item['proposer'] = self._proposer
                    proposals.put_item(Item=item)
                if handled_tag:
                    print("\n\n")

    @abc.abstractmethod
    def resources(self, session):
        raise NotImplementedError

    @abc.abstractmethod
    def handler(self, resource, expected_tags, region, session, cache, proposals):
        raise NotImplementedError



import sys
import time

from boto3 import Session
from botocore.exceptions import ClientError

from tag_a_day.cache import AWSCache
from tag_a_day.config import Configuration
from tag_a_day.log import logger
from tag_a_day.services import Services


def _initialise_common():
    def aws_session(region='us-east-1'):
        return Session(region_name=region)

    conf = Configuration(aws_session).parse()
    return aws_session, conf


def run():
    aws_session, conf = _initialise_common()
    cache = AWSCache(aws_session)

    table = aws_session(). \
        resource('dynamodb', conf.dynamodb_table_region). \
        Table(conf.dynamodb_table_name)
    services = Services(
        session=aws_session,
        cache=cache,
        services=conf.services,
        proposals=table)

    for region in conf.regions:
        session = aws_session(region)
        for service in services:
            logger().info(
                'Auditing tags for {0} in {1}'.format(service, region))
            service.run(
                expected_tags=conf.required_tags,
                region=region,
                session=session
            )


def initialise():
    aws_session, conf = _initialise_common()
    table_region = conf.dynamodb_table_region
    table_name = conf.dynamodb_table_name

    ddb_client = aws_session(table_region).client('dynamodb')
    print("Checking if table exists...")
    tableExists = True
    try:
        ddb_client.describe_table(TableName=table_name)
        tableExists = True
    except ClientError as e:
        exception = e.response.get('Error').get('Code')
        if exception == 'ResourceNotFoundException':
            tableExists = False

    if tableExists:
        print("Table '{name}' already exists.".format(
            name=table_name))
        sys.exit(1)

    table = aws_session(table_region).resource('dynamodb').create_table(
        TableName=table_name,
        KeySchema=[
            {
                'AttributeName': 'resource_id',
                'KeyType': 'HASH'
            },
            {
                'AttributeName': 'tag_key',
                'KeyType': 'RANGE'
            }
        ],
        AttributeDefinitions=[
            {
                'AttributeName': 'proposer',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'resource_type',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'resource_id',
                'AttributeType': 'S'
            },
            {
                'AttributeName': 'tag_key',
                'AttributeType': 'S'
            }
        ],
        ProvisionedThroughput={
            'ReadCapacityUnits': 5,
            'WriteCapacityUnits': 5
        },
        GlobalSecondaryIndexes=[
            {
                'IndexName': 'proposer-resource_type',
                'KeySchema': [
                    {
                        'AttributeName': 'proposer',
                        'KeyType': 'HASH'
                    },
                    {
                        'AttributeName': 'resource_type',
                        'KeyType': 'RANGE'
                    }
                ],
                'Projection': {
                    'ProjectionType': 'ALL'
                },
                'ProvisionedThroughput': {
                    'ReadCapacityUnits': 5,
                    'WriteCapacityUnits': 5
                }
            }
        ]
    )
    while table.table_status == 'CREATING':
        print("Table '{name}' is '{status}'...".format(
            status=table.table_status,
            name=table_name
        ))
        time.sleep(5)
        table.reload()

    print("Table '{arn}' has status '{status}'".format(
        arn=table.table_arn,
        status=table.table_status
    ))



from boto3.dynamodb.conditions import Key


class AWSCache(object):
    def __init__(self, session_function):
        self._session = lambda region: session_function(region)
        self._cache_store = {}

    def vpc(self, vpc_id, region):
        if vpc_id not in self._cache('vpc'):
            vpc = self._session(region).resource('ec2').Vpc(vpc_id)
            self._set_cache('vpc', vpc.id, vpc)
        return self._get_cache('vpc', vpc_id)

    def subnet(self, subnet_id, region):
        if subnet_id not in self._cache('subnet'):
            subnet = self._session(region).resource('ec2').Subnet(subnet_id)
            self._set_cache('subnet', subnet.id, subnet)
        return self._get_cache('subnet', subnet_id)

    def _cache(self, cache_id):
        if cache_id not in self._cache_store:
            self._cache_store[cache_id] = {}
        return self._cache_store[cache_id]

    def _set_cache(self, cache_id, object_id, cache_object):
        self._cache(cache_id)  # Ensure the cache exists
        self._cache_store[cache_id][object_id] = cache_object

    def _get_cache(self, cache_id, object_id):
        cache = self._cache(cache_id)
        return cache.get(object_id, None)


class ProgressCache(object):
    def __init__(self, table, proposer, resource_type):
        self._proposals = table
        self._proposer = proposer
        self._resource_type = resource_type
        self._progress_items = []

    def load(self):
        result = self._proposals.query(
            TableName=self._proposals.table_name,
            IndexName='proposer-resource_type',
            KeyConditionExpression=Key('proposer').eq(self._proposer) & Key('resource_type').eq(self._resource_type)
        )
        for item in result['Items']:
            del item['proposer']
            del item['resource_type']
            self._progress_items.append(item)
        return self

    def has_finished(self, resource_id, tag_names):
        evaluated_tags = self.evaluated_tags(resource_id)
        return set(tag_names) == set(evaluated_tags)

    def evaluated_tags(self, resource_id):
        return list(map(
            lambda r: r['tag_key'],
            filter(
                lambda r: r['resource_id'] == resource_id,
                self._progress_items)
        ))



from tag_a_day.services.service import Service


class RDSTagHandler(Service):
    name = 'rds_instance'
    missing_tags_text = "This instance is missing '{0}' in its tags"

    def resources(self, session):
        rds = session.client('rds')
        paginator = rds. \
            get_paginator('describe_db_instances'). \
            paginate()
        for instances_page in paginator:
            # Randomly pick 2/3rds of the nodes
            for instance in self._random_choose(instances_page['DBInstances']):
                instance['Tags'] = rds.list_tags_for_resource(
                    ResourceName=instance['DBInstanceArn']).get('TagList')
                yield instance

    def handler(self, instance, expected_tags, region, session, cache, proposals):
        # Get VPC info
        _, vpc_name, vpc = self._get_vpc_info(
            vpc_id=instance['DBSubnetGroup']['VpcId'],
            region=region
        )

        # Build table for displaying instance information
        evaluated_tags = self._progress.evaluated_tags(instance['DbiResourceId'])
        instance_info, missing_tags = \
            self._build_tag_sets(expected_tags, evaluated_tags, instance['Tags'])

        if self._progress.has_finished(instance['DbiResourceId'], expected_tags):
            self._skip(instance['DbiResourceId'])
            return

        if any(missing_tags):
            self._print_table(
                ("Vpc", vpc_name),
                ("VpcID", vpc.id),
                ("DatabaseARN", instance['DBInstanceArn']),
                *instance_info
            )

            tag_prompt = self._build_tag_prompt(missing_tags)
            for tag_key in missing_tags:
                yield {
                    'resource_id': instance['DbiResourceId'],
                    'tag_key': tag_key,
                    'tag_value': tag_prompt(tag_key),
                }



import io
from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with io.open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='aws-tag-a-day',
    version='0.1.5',
    packages=find_packages(exclude=('tests',)),
    description='A tool for simplifying swarming of fixing AWS tags',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/bliseng/aws-tag-a-day',
    license='Apache2',
    author='Drew J. Sonne',
    author_email='drew.sonne@gmail.com',
    install_requires=[
        'tabulate',
        'prompt_toolkit ',
        'hconf',
        'pyyaml',
        'boto3'
    ],
    setup_requires=["pytest-runner", "twine", "wheel"],
    tests_require=["pytest"],
    entry_points={
        'console_scripts': [
            'tag-a-day = tag_a_day.cli:run',
            'tag-a-day-initialise = tag_a_day.cli:initialise'
        ],
        'tag_a_day.tag_handlers': [
            'ec2 = tag_a_day.services.ec2:EC2TagHandler',
            'rds = tag_a_day.services.rds:RDSTagHandler',
            's3 = tag_a_day.services.s3:S3TagHandler',
            'emr = tag_a_day.services.emr:EMRTagHandler',
        ]
    }
)
